######################################
# File with Functions and Structures #
######################################

# Structure for parameters
mutable struct ParamsStruc{T <: Real}
    # Demand curve
    α::T        # Weight on leisure
    θ::T        # Capital share
    # Data
    τ_7074::Vector{T}        # Effective tax rate (1970-1974)
    τ_9396::Vector{T}        # Effective tax rate (1993-1996)
    CtoO_7074::Vector{T}     # Consumption to output ratio (1970-1974)
    CtoO_9396::Vector{T}     # Consumption to output ratio (1993-1996)
    h_7074::Vector{T}        # Hours (1970-1974)
    h_9396::Vector{T}        # Hours (1993-1996)
end

################################################################################

# Set parameter values
function SetParams(; α = 1.54, θ = 0.32,
                    τ_7074 = [0.52, 0.49, 0.41, 0.44, 0.45, 0.25, 0.40],
                    τ_9396 = [0.59, 0.59, 0.64, 0.52, 0.44, 0.37, 0.40],
                    CtoO_7074 = [0.66, 0.66, 0.66, 0.72, 0.77, 0.60, 0.74],
                    CtoO_9396 = [0.74, 0.74, 0.69, 0.77, 0.83, 0.68, 0.81],
                    h_7074 = [24.6, 24.4, 19.2, 22.2, 25.9, 29.8, 23.5],
                    h_9396 = [19.3, 17.5, 16.5, 22.9, 22.8, 27.0, 25.9])
    # Output
    ParamsStruc(α,θ,τ_7074,τ_9396,CtoO_7074,CtoO_9396,h_7074,h_9396)
end

################################################################################

# Computes hours worked using Prescott's formula
# It takes the vector of tax rates, tau, and the vector of
# consumption/output ratios, ctoo, as inputs and gives a vector for hours, h, as the output
function ComputeHours(α,τ,CtoO,h_data,Params)
    # Unpack parameters
    @unpack θ = Params
    # Hours from the model
    h = 100.0.*(1.0-θ)./( α.*CtoO./(1.0.-τ) .+ (1.0-θ) )
    # Prediction error
    error = h_data.-h
    # Switch sign for minimizing
    return h,error
end

################################################################################

# Output table with key statistics
function OutputTable(h_t1,h_t2,error_t1,error_t2,norm_error,Params)
    # Table header
    table_header = ["Data", "Model", "Prediction error"]
    table_data1 = [Params.h_7074  h_t1  error_t1]
    table_data2 = [Params.h_9396  h_t2  error_t2]
    # Print tables
    print("################################################################\n")
    print("Results: 1970-1974\n")
    pretty_table(table_data1; header = table_header, tf = tf_markdown, formatters = ft_printf("%5.2f", 2:3) )
    print("################################################################\n")
    print("Results: 1993-1996\n")
    pretty_table(table_data2; header = table_header, tf = tf_markdown, formatters = ft_printf("%5.2f", 2:3) )
    print("################################################################\n")
    print("Norm for the predition errors = ", norm_error, "\n")
    # Save table as dataframe
    df = DataFrame("Data (1970-74)" => Params.h_7074, "Data (1993-96)" => Params.h_9396,
                    "Model (1970-74)" => h_t1, "Model (1993-96)" => h_t2,
                    "Prediction error (1970-74)" => error_t1, "Prediction error (1993-96)" => error_t2)
    return df
end

################################################################################

# Specifies the first-order condition for the least-squares optimization problem
function FOC_α(α,Params)
    # Unpack parameters
    @unpack θ,CtoO_7074,CtoO_9396,τ_7074,τ_9396 = Params
    # Calculate the model's solution for hours worked for a given alpha
    h_model_9396 = 100.0.*(1.0-θ)./( α.*CtoO_9396./(1.0.-τ_9396) .+ (1.0-θ) )
    h_model_7074 = 100.0.*(1.0-θ)./( α.*CtoO_7074./(1.0.-τ_7074) .+ (1.0-θ) )
    # First-order condition for the minimization problem involving the choice of alpha
    FOC_resid = -(Params.h_9396 - h_model_9396).*Derivh(α,τ_9396,CtoO_9396,Params) -
                (Params.h_7074 - h_model_7074).*Derivh(α,τ_7074,CtoO_7074,Params)
    # Sum across the components of a vector
    FOC_resid = 2.0*sum(FOC_resid)
    # Function to be solved
    return FOC_resid
end

# Function to be solved for
FOC(x) = FOC_α(x,Params)

################################################################################

# Gives a vector for the derivative of hours worked with respect to α,
# given vectors of taxes, τ, and consumption to output ratios, CtoO
function Derivh(α,τ,CtoO,Params)
    # Unpack parameters
    @unpack θ = Params
    # Derivative of hours worked wrt α
    derivative = -100.0.*(1.0-θ)./( α*CtoO./(1.0.-τ) .+ (1.0-θ) ).^2 .*
                ( CtoO./(1.0.-τ) )
    # Output
    return derivative
end

################################################################################

# Specifies the objective function for the least-squares optimization problem
function SumOfSquares_α(α,Params)
    # Unpack parameters
    @unpack θ,h_7074,h_9396,CtoO_7074,CtoO_9396,τ_7074,τ_9396 = Params
    # Calculate the model's solution for hours worked for a given alpha
    h_model_9396 = 100.0.*(1.0-θ)./( α.*CtoO_9396./(1.0.-τ_9396) .+ (1.0-θ) )
    h_model_7074 = 100.0.*(1.0-θ)./( α.*CtoO_7074./(1.0.-τ_7074) .+ (1.0-θ) )
    # Sum of the squares for the minimization problem involving the choice of alpha
    SOS = (h_9396 - h_model_9396).^2.0 + (h_7074 - h_model_7074).^2.0
    # Sum across the components of a vector
    SOS = sum(SOS)
    # Output
    return SOS
end

# Function to be solved for
SumOfSquares(x) = SumOfSquares_α(x,Params)

################################################################################

function MakeFig(hdata, hmodel, heading)
    # Labels for points
    Labels = ["Ger", "Fra", "Ita", "Can", "UK", "Jap", "USA"]
    # Plot of 45 degree line
    p = plot(hdata, hdata,
            title = heading, titlefontsize=16,
            ylabel="Model", ytickfontsize=12, yguidefontsize=12,
            xlabel="Data", xtickfontsize=12,
            lw=3, legend=false)
    # Scatter plot of countries
    p = scatter!(hdata,hmodel,
                markershape = :diamond, markersize = 5, markercolor = :red,
                markerstrokecolor = :red,
                series_annotations = text.(Labels, 13, :bottom))
    # Save figure
    savefig("Figures/fig_$heading.pdf")
    return p
end

################################################################################
